// ModelSwarm.java
// The top-level ModelSwarm

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;

import swarm.objectbase.EmptyProbeMap;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.activity.ActionGroupImpl;
import swarm.activity.ScheduleImpl;
import swarm.activity.Schedule;
import swarm.activity.Activity;
import swarm.space.Grid2dImpl;
import swarm.space.Discrete2dImpl;
import swarm.collections.ListImpl;
import swarm.random.UniformIntegerDistImpl;



public class ModelSwarm extends SwarmImpl {
   
    // Declare the model parameters and their default values.
    
    public int worldXSize = 80, worldYSize = 80;
    public double seedProb = 0.80;
    public double bugDensity = 0.10;
    public int endTime = 125;

    public int bugHardiness=100;
    public int numberOfBugs=0;
    public int numTargets=0;
    public int startNumTargets=0;
    
    // Declare some other needed variables.
    
    public FoodSpace foodSpace;
    
    public Grid2dImpl bugSpace;
    public ListImpl bugList;
    
//    public TargetBug targetBug;
    public ListImpl targetList;
    public Grid2dImpl targetSpace;
    
    public ScheduleImpl modelSchedule;
    public Schedule createSchedule;
    public ListImpl reaperQueue;
 //   public EmptyProbeMapImpl probeMap;
    
    // This is the constructor for a new ModelSwarm. All we do is to
    // use the contructor for ModelSwarm's parent class.
    
    public ModelSwarm(Zone azone) {
        super(azone);

//let op: probeMap wordt niet als deel van deze class gedeclareerd!

    EmptyProbeMapImpl probeMap=new EmptyProbeMapImpl(azone, getClass());
    
    probeMap.addProbe(getProbeForVariable("worldXSize"));
    probeMap.addProbe(getProbeForVariable("worldYSize"));
    probeMap.addProbe(getProbeForVariable("seedProb"));
    probeMap.addProbe(getProbeForVariable("bugDensity"));
    probeMap.addProbe(getProbeForVariable("endTime"));
    
    Globals.env.probeLibrary.setProbeMap$For(probeMap,getClass());

}
 
    // Hier komen de accessors, vreemd genoeg (overgenomen van de 
    // ObjC code) voor build Objects()
    
    public final ListImpl getBugList() {
        return bugList;
    }
    
//    public final TargetBug getTargetBug() { return targetBug;}
    
    public final ListImpl getTargetList() { return targetList;}
    
    public final Grid2dImpl getTargetSpace() {return targetSpace; }
    
    public final Grid2dImpl getWorld() { return bugSpace; }
    
    public final FoodSpace getFood() { return foodSpace;  }

    public final int getWorldXSize() { return worldXSize; }
    public final int getWorldYSize() { return worldYSize; }
    
    
    // This is the method for building the model's objects: the food
    // space, the two-dimensional positioning grid, and the host of
    // bugs.
    
    public Object buildObjects() {
    
        int x, y, num;
        SimpleBug abug;
        TargetBug aTargetBug;
        
        // use the parent class buildObject() method to initialize the
        // process
        
        super.buildObjects();
        
        // Now create the model's objects.
        // First create the foodspace and seed it with food.
        
        foodSpace = new FoodSpace(Globals.env.globalZone,
        worldXSize, worldYSize);
        // foodSpace.seedFoodWithProb( seedProb ); we gebruiken geen 'vast' voedsel meer.
        
        bugSpace = new Grid2dImpl(Globals.env.globalZone, worldXSize, worldYSize);
        bugSpace.fastFillWithObject(null);
               
        targetSpace = new Grid2dImpl(Globals.env.globalZone, worldXSize, worldYSize);
        targetSpace.fastFillWithObject(null);
               
        bugList = new ListImpl(Globals.env.globalZone);
        reaperQueue = new ListImpl(Globals.env.globalZone);
        targetList = new ListImpl(Globals.env.globalZone);
                
      /*  num = 0;
        for (y = 0; y < worldYSize; y++)
            for (x = 0; x < worldXSize; x++)
                if ( Globals.env.uniformDblRand.getDoubleWithMin$withMax(0.0, 1.0) <= bugDensity/2) { 
                    abug = new SimpleBug(Globals.env.globalZone,bugSpace, x, y, ++num,
                         bugHardiness, this);
                    abug.setBugColor((byte)3); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    bugSpace.putObject$atX$Y(abug, x, y);
                    bugList.addLast(abug);
                }
             */
       
                    abug = new SimpleBug(Globals.env.globalZone,bugSpace, 3, 42, 1,
                         bugHardiness, this);
                    abug.archParam=1;
                    abug.setBugColor((byte)3); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    bugSpace.putObject$atX$Y(abug, 3, 42);
                    bugList.addLast(abug); 
                    
              /*      
                    abug = new SimpleBug(Globals.env.globalZone,bugSpace, 15, 33, 1,
                         bugHardiness, this);                                        
                    abug.archParam=1; //randomwalking bug
                    abug.setBugColor((byte)3); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    bugSpace.putObject$atX$Y(abug, 15, 33);
                    bugList.addLast(abug);
                    
                    abug = new SimpleBug(Globals.env.globalZone,bugSpace, 20, 28, 1,
                         bugHardiness, this);                                        
                    abug.archParam=1; //randomwalking bug
                    abug.setBugColor((byte)3); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    bugSpace.putObject$atX$Y(abug, 20, 28);
                    bugList.addLast(abug);
                    
                    abug = new SimpleBug(Globals.env.globalZone,bugSpace, 19, 3, 1,
                         bugHardiness, this);                                        
                    abug.archParam=1; //randomwalking bug
                    abug.setBugColor((byte)3); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    bugSpace.putObject$atX$Y(abug, 19, 3);
                    bugList.addLast(abug); */
                          
        num = 0;
        for (y = 0; y < worldYSize; y++)
            for (x = 0; x < worldXSize; x++)
                if ( Globals.env.uniformDblRand.getDoubleWithMin$withMax(0.0, 1.0) <= bugDensity*4) {
                    aTargetBug = new TargetBug(Globals.env.globalZone, targetSpace, x, y, 
					      Globals.env.uniformIntRand.getIntegerWithMin$withMax (1,4), ++num, this);
                    aTargetBug.setBugColor((byte)150); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
                    targetSpace.putObject$atX$Y(aTargetBug, x, y);
                    targetList.addLast(aTargetBug);
                } 

//        targetSpace.putObject$atX$Y(targetBug, targetBug.getXPos(), targetBug.getYPos());        
                
        System.out.println("We just created " + num + "bugs and put them in a List.");
        
        
        numberOfBugs=num;
        startNumTargets=num;
        
        // We're done.

        return this;
    }
    

/***************************************  BuildActions ****************************************/
    
    public Object buildActions()
        {
	    ActionGroupImpl modelActions;

        super.buildActions();

        modelActions = new ActionGroupImpl(getZone());
        createSchedule = new ScheduleImpl(this, 50);  // Create new agents once in 50 timesteps
                                                      // Only possible with ScheduleImpl and not with ActionGroupImpl

  /*    try{
        modelActions.createActionForEach$message
            (bugList, new Selector(Class.forName("SimpleBug"), "perceiveThinkMove",false));
      } catch(Exception e) {
        System.err.println("Exception in ModelSwarm sending perceiveAll: " +
            e.getMessage());
            System.exit(1);
      } */

       try{
        modelActions.createActionForEach$message
           (targetList, new Selector(Class.forName("TargetBug"), "uniformMotion", false));
       } catch(Exception e) {
            System.err.println("Exception sending message uniformMotion to TargetBug: " +
                e.getMessage());
            System.exit(1);
       } 
       
/*       try{
        modelActions.createActionForEach$message
            (bugList, new Selector(Class.forName("SimpleBug"), "updateQ", false));
       } catch(Exception e) {
            System.err.println("Exception sending message updateQ to bugList: " +
                e.getMessage());
            System.exit(1);
       } */

       try{
           modelActions.createActionTo$message
           (this, new Selector(this.getClass(), "reapBugs", false));
       } catch (Exception e)
       {
        System.err.println("Exception reapBugs: " +
            e.getMessage());
            System.exit(1);
       }
       
       try {
        createSchedule.at$createActionTo$message
        (1, this, new Selector(this.getClass(), "createAgents", false));
       } catch (Exception e) {
       System.err.println("Exception in createAgents: " + e.getMessage());
       System.exit(1);
       }
            
        
        modelSchedule = new ScheduleImpl(getZone(), 1);       
        modelSchedule.at$createAction(0, modelActions);

        return this;
    }
    
    // This method specifies the context in which the model is to be run.
    
    public Activity activateIn(Swarm swarmContext) {
            
        // Use the parent class to activate ourselves in the context
        // passed to us.
    
        super.activateIn(swarmContext);
        
        // Then activate the schedule in ourselves.
        
        modelSchedule.activateIn(this);
        createSchedule.activateIn(this);

        
        // Finally, return the activity we have built.
        
        return getActivity();
    }

/*************************************  reportInterception: agent runs into a target ************************************/

    public void reportInterception(int xInterceptPos, int yInterceptPos) {
        TargetBug targetToBeRemoved;


        ((SimpleBug)bugSpace.getObjectAtX$Y(xInterceptPos,yInterceptPos)).notifyInterception();

  
        targetToBeRemoved=(TargetBug)targetSpace.getObjectAtX$Y(xInterceptPos,yInterceptPos);
        System.out.println("Target to be removed: " + targetToBeRemoved.agentNumber + " pos: " + targetToBeRemoved.xPos + ", " + targetToBeRemoved.yPos);
        //targetList.remove(targetToBeRemoved);
        targetSpace.putObject$atX$Y(null, xInterceptPos, yInterceptPos);
        reaperQueue.addLast(targetToBeRemoved);
    }

/*************************************  removeTarget: target disappears   **********************************************/

    public void removeTarget(TargetBug target) {
        if(target != null) {
            targetSpace.putObject$atX$Y(null, target.xPos, target.yPos);
            reaperQueue.addLast(target);
        }
    }


    public void bugDeath(SimpleBug abug)
    {
	if ((SimpleBug)bugSpace.getObjectAtX$Y(abug.xPos, abug.yPos) == abug)
	    {
	    bugSpace.putObject$atX$Y(null, abug.xPos, abug.yPos);
	    reaperQueue.addLast(abug);
	    }
    }
public void reapBugs()
    {
	Agent abug;

	while (reaperQueue.getCount() != 0)
	    {        
	    if(reaperQueue.getFirst() instanceof SimpleBug) {
	        abug=(SimpleBug)reaperQueue.removeFirst();
	        bugList.remove(abug);
	        abug.drop();
	        System.out.println("Bug " + abug.agentNumber + 
			        " has died of hunger.");
	        numberOfBugs -= 1;
	    } 
	    if((reaperQueue.getCount()!=0) && (reaperQueue.getFirst() instanceof TargetBug)) {
	        abug=(TargetBug)reaperQueue.removeFirst();
	        targetList.remove(abug);
	        abug.drop();
	        System.out.println("Target " + abug.agentNumber + 
			        " has been removed.");
	        numTargets -= 1;
	    }
	    
	    }  
	if (bugList.getCount() == 0)	    
	    getActivity().terminate();

    }

public void createAgents() {
    
    int newX, newY, varNumTar;
 
    System.out.println("***************************************");
    System.out.println("NOW IN createAgents");
    System.out.println("***************************************");

    varNumTar=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-10, 10);
  
    while(targetList.getCount() < startNumTargets + varNumTar) { 
            
        newX=Globals.env.uniformIntRand.getIntegerWithMin$withMax(0, worldXSize-1);
        newY=Globals.env.uniformIntRand.getIntegerWithMin$withMax(0, worldYSize-1);
        
        // Test if there is no target at this spot, one left, right, under or above (no diagonals tested).
        
        if((newX<0)||(newY<0)||(((newX-1+worldXSize) % worldXSize) <0) || (((newY-1+worldYSize) % worldYSize) < 0)){
            System.out.println("!!!!!!!!!! Coordinates out of bound! (" + newX + ", " + newY + ") !!!!!!!!!!!!!!!!");
            System.exit(1);
        }
        
        if ((targetSpace.getObjectAtX$Y(newX, newY)==null) &&
            (targetSpace.getObjectAtX$Y(((newX + 1 + worldXSize) % worldXSize), newY)==null) &&
            (targetSpace.getObjectAtX$Y(((newX -1 + worldXSize) % worldXSize), newY)==null) &&
            (targetSpace.getObjectAtX$Y(newX,((newY + 1 + worldYSize)% worldYSize))==null) &&
            (targetSpace.getObjectAtX$Y(newX,((newY - 1 + worldYSize) % worldYSize))==null)) {

            TargetBug aTargetBug = new TargetBug(Globals.env.globalZone, targetSpace, newX, newY, 
						  Globals.env.uniformIntRand.getIntegerWithMin$withMax(1,4), targetList.getCount()+1, this);
            aTargetBug.setBugColor((byte)150); // dit is groen, maar het zou eigenlijk al met het setten van de colormap gebeurd moeten zijn.
            targetSpace.putObject$atX$Y(aTargetBug, newX, newY);
            targetList.addLast(aTargetBug);
            
            System.out.println("A new TargetBug created at: (" + newX + ", " + newY + ")");
            
        }
    }
    
    printLengthAllLists();

}

    public void printLengthAllLists() {
        System.out.println("The length of ListImpl bugList is: " + bugList.getCount());
        System.out.println("The length of ListImpl targetList is: " + targetList.getCount());
        System.out.println("The length of ListImpl reaperQueue is: " + reaperQueue.getCount());

        System.out.println("The length of ListImpl bug1.currentlyPerceivedTargets is: " + (((SimpleBug)bugList.getFirst()).currentlyPerceivedTargets).getCount());
        System.out.println("The length of ListImpl bug2.currentlyPerceivedTargets is: " + (((SimpleBug)bugList.getLast()).currentlyPerceivedTargets).getCount());

        System.out.println("The length of ListImpl bug1.perceivedTargets is: " + (((SimpleBug)bugList.getFirst()).perceivedTargets).getCount());
        System.out.println("The length of ListImpl bug2.perceivedTargets is: " + (((SimpleBug)bugList.getLast()).perceivedTargets).getCount());

        System.out.println("The length of ListImpl bug1.epiPos is: " + (((SimpleBug)bugList.getFirst()).epiPos).size());
        System.out.println("The length of ListImpl bug2.epiPos is: " + (((SimpleBug)bugList.getLast()).epiPos).size());
        
        

    }
    
    // This is a pretty crude method to end the simulation after an
    // arbitrary number of periods given by the endTime parameter. If
    // the simulation time returned by getCurrentTime() is greater
    // than endTime, we send a message to the console and terminate
    // the current activity. We also use this method to send a
    // timestamp to the console.
    public void checkTime() {
        int i, t;
        int interval = 5;
        if (Globals.env.getCurrentTime() >= endTime){
            
            // Terminate the simulation.
            
            System.out.println("We've reached our endTime at period " +
            Globals.env.getCurrentTime());
            Globals.env.getCurrentSwarmActivity().terminate();
        }
        else{
            for (i = 0; (t = (int)Math.pow(interval, (double)i)) <= endTime; i++)
                if (t == Globals.env.getCurrentTime()) 
                    System.out.println("The time is " + t);
        }
        return;
    }
	//{{DECLARE_CONTROLS
	//}}
}
