// Java Heatbugs application. Copyright  1999-2000 Swarm Development Group.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular
// purpose.  See file COPYING for details and terms of copying.

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.ZoneImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;
import swarm.objectbase.ProbeMap;

import swarm.collections.ListImpl;


import swarm.gui.Colormap;
import swarm.gui.ColormapImpl;
import swarm.gui.ZoomRaster;
import swarm.gui.ZoomRasterImpl;

import swarm.analysis.EZGraph;
import swarm.analysis.EZGraphImpl;

import swarm.simtoolsgui.GUISwarm;
import swarm.simtoolsgui.GUISwarmImpl;

import swarm.space.Value2dDisplay;
import swarm.space.Value2dDisplayImpl;
import swarm.space.Object2dDisplay;
import swarm.space.Object2dDisplayImpl;

import java.util.List;

public class ObserverSwarm extends GUISwarmImpl {

//  public ParameterManager parameterManager;

  public ActionGroup displayActions;				
  public Schedule displaySchedule;
  public ModelSwarm modelSwarm;	  	
  public int displayFrequency;
  public int zoomFactor=7;
  public int simulatedTime=0;
  
  // The display objects
  
  public Colormap colorMap;
  public ZoomRaster worldRaster;
  public Value2dDisplay foodDisplay;
  public Object2dDisplay bugDisplay;
  public Object2dDisplay targetDisplay;
  public EZGraphImpl resultGraph;

  public ObserverSwarm (Zone aZone) {
    super(aZone);
    displayFrequency = 1;
    
  }
  
/********************************** buildObjects ****************************/

  public Object buildObjects () {
  
  super.buildObjects ();

  Zone parameterZone = new ZoneImpl(getZone());

  modelSwarm = (ModelSwarm) Globals.env.lispAppArchiver.getWithZone$key
                                                    (Globals.env.globalZone, "modelSwarm");
  
	Globals.env.createArchivedProbeDisplay(modelSwarm, "modelSwarm");
	Globals.env.createArchivedProbeDisplay(this, "observerSwarm");
  
  
  getControlPanel ().setStateStopped ();
        
  modelSwarm.buildObjects ();
        
  //parameterManager=(ParameterManager) Globals.env.lispAppArchiver.getWithZone$key
  //                                                  (parameterZone, "parameterManager");
        
  // Na het bouwen van de ModelSwarm Objecten nu die van Observer Swarm
  
  /* Omdat we geen Class ParameterManager kunnen bouwen vanwege oude viscafe versie, moeten
     we bugHardiness halen uit een bug. Erg omslachtig. Let op de casts die worden uitgevoerd.
     Dat is nodig want anders ziet Java het als Java Objects die worden geretourneerd.*/
    
  int bugHardiness = ((SimpleBug)((ListImpl)modelSwarm.getBugList()).getFirst()).getBugHardiness();
  
  System.out.println("De bugHardiness zoals die verkregen is in de ObserverSwarm is: " + bugHardiness);
  
  
  colorMap= new ColormapImpl(getZone());
    
  int colornumber=0;
  
 
  colorMap.setColor$ToName((byte)0,"black");
  colorMap.setColor$ToName((byte)1,"red");
  colorMap.setColor$ToName((byte)2,"yellow");
  colorMap.setColor$ToName((byte)3,"green");
  colorMap.setColor$ToName((byte)150,"blue");
 
  for(int i=0; i<bugHardiness;i++)
    colorMap.setColor$ToRed$Green$Blue((byte)(i+4), 
        (double) 0.5, //percentage rood: hoe hoger het kleurnummer hoe minder rood
         (double) ( ((double)i/bugHardiness) *.8+.2),     //percentage groen: van 20 tot 100 LET OP: i is een integer!
                  0);

   
  worldRaster = new ZoomRasterImpl(getZone(),"worldRaster");
  worldRaster.setColormap(colorMap);
  worldRaster.setZoomFactor(4);
  
  worldRaster.setWidth$Height
    ((modelSwarm.getWorld()).getSizeX(),
    (modelSwarm.getWorld()).getSizeY());
 
  worldRaster.setWindowTitle("Food World");
  worldRaster.pack();
          
  
  foodDisplay = new Value2dDisplayImpl
    (getZone(), worldRaster, colorMap, modelSwarm.getFood());
 
  try{
    bugDisplay = new Object2dDisplayImpl
        (getZone(), worldRaster, modelSwarm.getWorld(),
         new Selector(Class.forName("SimpleBug"),"drawSelfOn", false));
  } catch (Exception e) {
    System.err.println("Exception in buildObjects drawSelfOn: " + e.getMessage());
  }
  
  bugDisplay.setObjectCollection
    (modelSwarm.getBugList());
 
  try{
    targetDisplay = new Object2dDisplayImpl
        (getZone(), worldRaster, modelSwarm.getTargetSpace(),
        new Selector(Class.forName("TargetBug"),"drawSelfOn", false));
  }catch (Exception e) {
    System.err.println("Exception in buildObjects for TargetBug.drawSelfOn: " + e.getMessage());
  }
   
  targetDisplay.setObjectCollection(modelSwarm.getTargetList());      
   
    // Maak een probe als er met de rechterknop geklikt wordt op het worldRaster 

    try{
    worldRaster.setButton$Client$Message(1,bugDisplay,  // Voor de hunters
        new Selector(bugDisplay.getClass(),"makeProbeAtX$Y",false));
    } catch(Exception e) {
        System.err.println("Exception in buildObjects makeProbeAtX$Y: " + e.getMessage());
    }
    
    try{
    worldRaster.setButton$Client$Message(3,targetDisplay, // Voor de prey
        new Selector(targetDisplay.getClass(),"makeProbeAtX$Y",false));
    } catch(Exception e) {
        System.err.println("Exception in buildObjects makeProbeAtX$Y (targetDisplay): " + e.getMessage());
    }
    
    // We are putting a probe on bug #1 to track its behaviour from scratch. Ik toegevoegd.
    
    ListImpl bugList = modelSwarm.getBugList();
    ListImpl targetList = modelSwarm.getTargetList();
  
    
    SimpleBug targetBug;
    
  /*  for(int i=0; i < bugList.getCount();i++) {
        if( ( ((targetBug = (SimpleBug)(bugList.atOffset(i))).getAgentNumber()) == 1) ) {
            Globals.env.createProbeDisplay(bugList.atOffset(i));
        }
            //System.out.println("In MakeObjects Bug #1 gevonden!");
    } */

   resultGraph = new EZGraphImpl(getZone(), "Periods Since Eaten", "Time Steps", 
                    "Periods", "resultGraph");

   try{
    resultGraph.createSequence$withFeedFrom$andSelector("maxWAPSE QBug",  bugList.getFirst(),
        new Selector(Class.forName("SimpleBug"), "getMaxWAPSE", false));
   } catch(Exception e) {
    System.err.println("Exception in buildObjects resultGraph.createSequence$withFeedFrom$andSelector: " + e.getMessage());
   }

  /* try{
    resultGraph.createSequence$withFeedFrom$andSelector("maxWAPSE RBug",  bugList.atOffset(1),
        new Selector(Class.forName("SimpleBug"), "getMaxWAPSE", false));
   } catch(Exception e) {
    System.err.println("Exception in buildObjects resultGraph.createSequence$withFeedFrom$andSelector: " + e.getMessage());
   } */
   
/*   TargetBug targetTBug;
   
     for(int i=0; i < targetList.getCount();i++) {
        if( ( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 11)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 12)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 13)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 14)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 15)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 16)
             ||( ((targetTBug = (TargetBug)(targetList.atOffset(i))).getAgentNumber()) == 17)
             ) {
            Globals.env.createProbeDisplay(targetList.atOffset(i));
  
        }   
     }
  */  
  return this;
  }  
    
/******************************* _update ******************************/    
    
  /* In de ObjC versie staan de statements in _update_ in buildActions,
     maar in jHeatBugs in deze afzonderlijke method, vandaar hier ook */
    
  public Object _update_() {
    if(worldRaster != null) {
        foodDisplay.display();
        bugDisplay.display();
        targetDisplay.display(); // hier gaat het fout 22/3/01
        worldRaster.drawSelf();
    }
    else System.err.println("Diplay in _update_ unsuccessful due to null pointer of worldRaster");
    return this;
  }
  
public Object showStats(){
    resultGraph.step();
    return this;
}
  
    
/*****************************  BUILDACTIONS *********************************/    
    
  public Object buildActions () {
    
    super.buildActions();
        
    modelSwarm.buildActions();   
  
  // Vanaf hier gaan we display acties invoeren.
  // Afgekeken van simpleObserverBug (Objc) en jHeatbugs

  displayActions = new ActionGroupImpl(getZone()); 
  // Don't forget this! The compiler won't complain if left out. A null pointer
  // exception will be raised throught the catch block below.
  
  
 try {
    displayActions.createActionTo$message
        (this, 
         new Selector(getClass(),"updateSimulatedTime", false));
 }
 catch(Exception e) {
    System.err.println("Exception in messaging ObserverSwarm with updateSimulatedTime: " + e.getMessage());
 }

  try {
    displayActions.createActionTo$message
        (this, new Selector(getClass(),"_update_",false));
  } catch (Exception e) {
    System.err.println("Exception in messaging _update_: " + e.getMessage());
    } 

  try {
    displayActions.createActionTo$message
        (this, new Selector(getClass(),"showStats",false));
  } catch (Exception e) {
    System.err.println("Exception in messaging showStats: " + e.getMessage());
  }
  
  try {
    displayActions.createActionTo$message
        (Globals.env.probeDisplayManager,
         new Selector(Globals.env.probeDisplayManager.getClass(), "update", false));
  }
  catch (Exception e){
    System.err.println("Exception in messaging probeDisplayManager: " + e.getMessage());
  }

  
   try {
    displayActions.createActionTo$message(
        getActionCache(), 
        new Selector (getActionCache().getClass(), "doTkEvents", true));
  } catch(Exception e) {
    System.err.println("Exception in setting up displayActions: " + e.getMessage());
    } 
  
  displaySchedule = new ScheduleImpl (getZone(), displayFrequency);
  displaySchedule.at$createAction(0,displayActions);
  
    return this;
  }  
  
  public Activity activateIn (Swarm swarmContext) {
    super.activateIn (swarmContext);
    modelSwarm.activateIn (this);
    displaySchedule.activateIn(this); // Moet er ook bij om het control panel te activeren.
    return getActivity();
  }

public void updateSimulatedTime() {
    simulatedTime=Globals.env.getCurrentTime();
}

}
    
  
