// SimpleBug.java for SimpleSwarmBug2
// Defines the class for our SimpleBug agents/

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import swarm.space.Grid2dImpl;
import swarm.gui.Raster;

import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;
import swarm.objectbase.ProbeMap;
import swarm.objectbase.ProbeLibrary;
import swarm.objectbase.ProbeLibraryImpl;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.random.UniformDoubleDistImpl;

import swarm.collections.ListImpl;
import swarm.collections.MapImpl;

import java.lang.Math;
import java.util.LinkedList;

public class SimpleBug extends Agent
{

//new variables

public int archParam=0; // parameter to switch between architectures of the agents.
public int bugHardiness;
public byte bugColor;
public boolean haveEaten;
public int periodsSinceEaten=0;
public int timesEaten=0;
public double maxWAPSE; // walking average of total periods since eaten when having eaten
public int perceptRadius=3;
public int perceptXDim=perceptRadius*2+1;
public int perceptYDim=perceptRadius*2+1;
public MapImpl currentlyPerceivedTargets;

public TarWAttrib lockedTarget=null;
public int episodeStatus;  //  0: not in an episode,   1: in an episode
public int updateQStatus;
public double[][][][] rlGrid;
public MapImpl perceivedTargets = new MapImpl(this.getZone());
public LinkedList epiPos;


/******************************* class tarWAttrib *********************************/
// Class to store the target with attributes, among which relative position 
// (it's a container, not an extention). 
// It is the target as it is perceived by the agent. Note that the Target
// is explicitly contained (the pointer to it) in this class

class TarWAttrib {
    TargetBug percTarget;
    int relXPosTar;
    int relYPosTar;
    
    TarWAttrib(TargetBug t, int relXP, int relYP) {
        percTarget=t;
        relXPosTar=relXP;
        relYPosTar=relYP;
    }
    
    
    public final TargetBug getTarget() { return percTarget; } 
}

/*******************************  class AgTarPos ***********************************/
/* This is a datastructure to store the conjunction of a place of the agent and the*/
/* Target in order to update after an episode has ended                             */

 class AgTarPos {
     int tarNum;
     int xPosAg;
     int yPosAg;
     int dirAg;
     int xPosTar;
     int yPosTar;
     int dirTar;
     int relYPosTar;
     int relXPosTar;

    AgTarPos(int tarN, int xPAg, int yPAg, int dAg, int xPTar, int yPTar, int dTar) {
        this.tarNum=tarN;
        this.xPosAg=xPAg;
        this.yPosAg=yPAg;
        this.dirAg=dAg;
        this.xPosTar=xPTar;
        this.yPosTar=yPTar;
        this.dirTar=dTar;
        this.relXPosTar=-1;
        this.relYPosTar=-1;
    }
    
    AgTarPos(int tarN, int xPAg, int yPAg, int dAg, int xPTar, int yPTar, int dTar, int rXPT, int rYPT) {
        this.tarNum=tarN;
        this.xPosAg=xPAg;
        this.yPosAg=yPAg;
        this.dirAg=dAg;
        this.xPosTar=xPTar;
        this.yPosTar=yPTar;
        this.dirTar=dTar;
        this.relXPosTar=rXPT;
        this.relYPosTar=rYPT;
    }

     void printPos() {
        System.out.println("A(" + xPosAg + ", " + yPosAg + ")R:" + dirAg + " T#" + tarNum + "(" + xPosTar + 
            ", " + yPosTar + ") Rel: (" + relXPosTar + ", " + relYPosTar + ")R:" + dirTar +
            " Q:" + rlGrid[dirAg][relXPosTar][relYPosTar][dirTar]);

     }
     
     final void setAgDir(int dir) { dirAg = dir; }
    
}


/***********************************************************************************/

/****************************  Constructor *****************************************/

    // Constructor to create a SimpleBug object in Zone aZone and to
    // place it in the foodspace and bugspace, fSpace and bSpace, at
    // the specified X,Y location. The bug is also given a numeric id,
    // bNum.

public SimpleBug(Zone aZone, Grid2dImpl bSpace,int X, int Y, int bNum,
                    int bHardiness, ModelSwarm mSwarm) {

    super(aZone, X, Y, 2, bNum, mSwarm, bSpace, (byte)0);

    myBugSpace = bSpace;
    bugHardiness = bHardiness;
   
  rlGrid = new double[directions][perceptXDim][perceptYDim][directions];  
  currentlyPerceivedTargets = new MapImpl(this.getZone());
  epiPos = new LinkedList();
  EmptyProbeMapImpl probeMapSimpleBug = new EmptyProbeMapImpl(aZone, getClass()); 

  probeMapSimpleBug.addProbe(getProbeForVariable("bugColor"));
  probeMapSimpleBug.addProbe(getProbeForVariable("periodsSinceEaten"));
  probeMapSimpleBug.addProbe(getProbeForVariable("maxWAPSE"));

  probeMapSimpleBug.addProbe(probeVariable("xPos"));
  probeMapSimpleBug.addProbe(probeVariable("yPos"));
  probeMapSimpleBug.addProbe(probeVariable("agentNumber"));
  
  
  Globals.env.probeLibrary.setProbeMap$For(probeMapSimpleBug,getClass());
 

}

/*************************************** Accessors *********************************************/

//ACCESSORS specific for this class

final public int getBugHardiness(){ return bugHardiness; }
final public double getMaxWAPSE() { return maxWAPSE; }

VarProbe probeVariable (String name)
{
    VarProbe varProbe = null;
    Class baseClass = super.getClass();

    do {
        varProbe =
            Globals.env.probeLibrary.getProbeForVariable$inClass
            (name, baseClass);
        baseClass = baseClass.getSuperclass ();
    } while (varProbe == null && baseClass != null);
    return varProbe;
} 

/****************************** perceiveThinkMove ***********************************************/
/*                                                                                              */
/* Method to let the agent look around itself and to determine its action and episodestatus     */
/*                                                                                              */
/************************************************************************************************/


public void perceiveThinkMove() {
    
    TargetBug tmpTarget;
   
// Positions relative to the target or in the perceptionfield. This feeds into the parametermatrix. 
// The coordinate system is the same as the grid2dImpl: The left upper hand is (0,0), right and down is counting up.
    int xRelWatch;
    int yRelWatch;
    
    
// currentlyPerceivedTargets had to be made a class variable. It needs to be available throughout the entire episode,
// _also_ at the end when interception takes place. Therefore, it needs to be cleaned up at the _start_ of each episode.

currentlyPerceivedTargets.deleteAll();
 
    if (agentNumber==1) {
        for(int i=0;i<=perceptRadius*2;i++){
            for(int j=0;j<=perceptRadius*2;j++){
                int xWatch=(xPos-perceptRadius+j+worldXSize) % worldXSize;
                int yWatch=(yPos-perceptRadius+i+worldYSize) % worldYSize;
                if((tmpTarget=(TargetBug)(modelSwarm.getTargetSpace()).getObjectAtX$Y(xWatch,yWatch)) != null) {

                        currentlyPerceivedTargets.at$insert(new Integer(tmpTarget.getAgentNumber()), 
                                new TarWAttrib(tmpTarget,i,j));
                   
                        
                 }                
            }
        }

        // Main condition: there is something in the perception buffer. Then switch on the episodeStatus.

        if(currentlyPerceivedTargets.getCount() != 0) {
           
            switch(episodeStatus) {
                case 0 : 
                    {   episodeStatus=1;
                        updateQStatus=1;
                        lockedTarget = (TarWAttrib)currentlyPerceivedTargets.getFirst();
                        
             // Record the position of the agent and target in the episode list.
             // The list should be empty. If not there's something wrong with the episode bookkeeping.
                        if (epiPos.size()!= 0) {
                            System.err.println("Error in perceiveThinkMove: episode list not empty while episode status is 0. Exiting.");
                            System.exit(1);
                        }
                        else {
                            epiPos.addLast(new AgTarPos(lockedTarget.percTarget.getAgentNumber(), xPos, yPos, direction,
                                                lockedTarget.percTarget.getXPos(), lockedTarget.percTarget.getYPos(), lockedTarget.percTarget.getDirection(),
                                                lockedTarget.relXPosTar, lockedTarget.relYPosTar));
                        }
                    }; 
                    break; 

                case 1 : 
                    {   TarWAttrib tmpWATar = (TarWAttrib)currentlyPerceivedTargets.at(new Integer(lockedTarget.percTarget.getAgentNumber()));
                        if ( tmpWATar!=null ){                            
                            episodeStatus=1;
                            updateQStatus=1;
                            epiPos.addLast(new AgTarPos(lockedTarget.percTarget.getAgentNumber(), xPos, yPos, direction,
                                                lockedTarget.percTarget.getXPos(), lockedTarget.percTarget.getYPos(), lockedTarget.percTarget.getDirection(),
                                                tmpWATar.relXPosTar, tmpWATar.relYPosTar));
                            lockedTarget.relXPosTar=tmpWATar.relXPosTar;
                            lockedTarget.relYPosTar=tmpWATar.relYPosTar;
                        }
                        else {  // The locked target is no longer perceived.
                            episodeStatus=1;
                            updateQStatus=3;   
                            updateQ();
                       
                            lockedTarget=(TarWAttrib)currentlyPerceivedTargets.getFirst(); 
                        
                        
                        
                        // Here the update procedure must take place and the previous episode list be purged.
                        epiPos.clear();
                            epiPos.addLast(new AgTarPos(lockedTarget.percTarget.getAgentNumber(), xPos, yPos, direction,
                                                lockedTarget.percTarget.getXPos(), lockedTarget.percTarget.getYPos(), lockedTarget.percTarget.getDirection(),
                                                lockedTarget.relXPosTar, lockedTarget.relYPosTar));
                        }
                    };
                    break;
            }
        }
   
        /* Below the case in which no targets are perceived. The status shifts until 0 is reached. */
        else {
            switch(episodeStatus) {
                case 0 : { 
                    updateQStatus=0;
                    episodeStatus=0; 
                }
                break; // No Target previously in view and no-one now.
                
                case 1 : {   
                    // Target was in view in previous timestep but now no more.
                    // Call updateQ to end episode just as above but now with no new targets in the perceptionbuffer.
                        updateQStatus=3;
                        updateQ();
                        epiPos.clear();
                        episodeStatus=0;
                        lockedTarget=null;
                 };
                 break;
            };
        }

//        perceivedTargets.deleteAll();                 // Dit gaat allemaal fout
//        perceivedTargets=new MapImpl(this.getZone());
//        currentlyPerceivedTargets.deleteAll();
        perceivedTargets=currentlyPerceivedTargets;
        

    }
    if (archParam==0) randomWalk();    
    else reallyWalkRandomly();

}

public void reportPosition() {
    System.out.println("Bug " + agentNumber + " is at " + xPos + ", " + yPos);
}

/*************************************************  randomWalk ***************************************/

public void randomWalk(){

    int tmpRandX;
    int tmpRandY;
    double[] newQ = new double[directions];


    // Q-value calculation
 
    if(episodeStatus>0){
    
        for(int i=0;i<directions;i++) {
            newQ[i]=rlGrid[i][lockedTarget.relXPosTar][lockedTarget.relYPosTar][lockedTarget.percTarget.direction];
        }
    }

    int newX, newY, direction;
    double tmpQ;
    double tmpRand[]= new double[directions];
    
    for(int i=0;i<directions;i++)
        tmpRand[i]= Globals.env.uniformDblRand.getDoubleWithMin$withMax(-0.25, 0.25);
    
    if(newQ[0]+tmpRand[0]>newQ[1]){  // Action 0: NOP
        tmpQ=newQ[0];
        direction=0;
        newX=xPos;
        newY=yPos;
    }
    else {                          // Action 1: North
        tmpQ=newQ[1];     
        direction=1;
        newY=yPos-1;
        newX=xPos;
    }
    if(tmpQ+tmpRand[1]<newQ[2]){    // Action 2: South
        tmpQ=newQ[2];
        direction=2;
        newX=xPos;
        newY=yPos+1;
    }
    if(tmpQ+tmpRand[2]<newQ[3]){    // Action 3: East
        tmpQ=newQ[3];
        direction=3;
        newX=xPos+1;
        newY=yPos;
    }
    if(tmpQ+tmpRand[3]<newQ[4]){    // Action 4: West
        direction=4;
        newX=xPos-1;
        newY=yPos;
    }
        
    newX = (newX + worldXSize) % worldXSize;
    newY = (newY + worldYSize) % worldYSize;
       
    //Here the action choice of the agent is recorded into the episode track.
    //The agent should currently be in an episode (i.e. episodeStatus > 0 otherwise 
    //there should be nothing recorded.
    
    if(episodeStatus>0) ((AgTarPos)epiPos.getLast()).dirAg=direction;
    
    // Is there a bug at the new position already? If not, put a
    // null at this bug's current position and put this bug at the
    // new position.
    /* hier heeft bug 1 gescand, maar nog geen stap gezet. houd in de gaten of dit zo moet zijn!
       de volgorde moet zijn: perceptie->actie. Wat er bij het debuggen aan de hand is, is dat de
       scan gerapporteerd wordt aan de console _voordat_ de simulatiestap voltooid is. Dat betekent dat
       in de probe er een andere positie wordt gerapporteerd dan na de scan procedure zoals die hier staat. */
  
    
    
    if (myBugSpace.getObjectAtX$Y(newX, newY) == null){  // Hier stond ipv newX xPos en newY yPos
        myBugSpace.putObject$atX$Y(null, xPos, yPos);
        xPos = newX;
        yPos = newY;
        myBugSpace.putObject$atX$Y(this, newX, newY);
        }
        
/* The eat procedure has been moved to notifyInterception, the haveEaten flag is always reset here. 
    The question is now, if there is a condition in which the agent has its stomach full... */

    haveEaten=false;
    ++periodsSinceEaten;
    setBugColor((byte)(bugHardiness-periodsSinceEaten+4));

    if (periodsSinceEaten >= bugHardiness*5)
        super.modelSwarm.bugDeath(this);
    }

public void reallyWalkRandomly(){
    int newX, newY;
    // Turn off the haveEaten flag.
    haveEaten = false;
    // Decide where to move.
    newX = xPos +
    Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
    newY = yPos +
    Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
    newX = (newX + worldXSize) % worldXSize;
    newY = (newY + worldYSize) % worldYSize;
    
    // Is there a bug at the new position already? If not, put a
    // null at this bug's current position and put this bug at the
    // new position.
    
    if (myBugSpace.getObjectAtX$Y(newX, newY) == null){
        myBugSpace.putObject$atX$Y(null, xPos, yPos);
        xPos = newX;
        yPos = newY;
        myBugSpace.putObject$atX$Y(this, xPos, yPos);
        }
        
    // If there is food at this cell, eat it and set the haveEaten
    // flag.
    
    haveEaten=false;
    ++periodsSinceEaten;
    setBugColor((byte)(bugHardiness-periodsSinceEaten+4));

    if (periodsSinceEaten >= bugHardiness*5)
        super.modelSwarm.bugDeath(this);
    
    
    }




/****************************************** notifyInterception ****************************************/
/*  Here the Agent is notified of the interception. This depends _critically_ on the activity schedule.
    The Agent moves first, then the target moves, and the target reports to modelswarm, which, in turn, 
    notifies the Agent and deletes the target.
    Also, it is now _here_ that the episode status changes and that the Agent eats, instead of in the
    movement procedure. */

public void notifyInterception() {
 
    /*  Before the interception is actually acted on a safety check takes place to ensure that there is
        indeed a Target at the place of the Agent. Else the sim exits because this is a serious error. */

    Object aTarget;
 
    if ((aTarget=modelSwarm.getTargetSpace().getObjectAtX$Y(xPos, yPos)) instanceof TargetBug){
    
       // System.out.println("Bug #" + agentNumber + " notified of interception at " + xPos + ", " + yPos);
        
            haveEaten = true;
            ++timesEaten;
            maxWAPSE=(maxWAPSE*(timesEaten-1)+periodsSinceEaten)/timesEaten;
            periodsSinceEaten = 0;
            if(agentNumber==1) 
                setBugColor((byte)2);
            else setBugColor((byte)2);

//If the target eaten is the chased target, i.e., the one with a lock, end the episode and update
//Set episodeStatus to 0 to indicate that the current episode has ended.
        if((archParam==0)&&(aTarget==lockedTarget.percTarget)) {
            
            TarWAttrib tmpTWAt = (TarWAttrib)currentlyPerceivedTargets.at(new Integer(lockedTarget.percTarget.getAgentNumber()));

            // If interception has taken place, the episode ends _here_ (instead of in the 
            // beginning of the timecycle) and the newly taken position must be stored in
            // the tracklist (epiPos) _before_ updating the parameters.
            epiPos.addLast(new AgTarPos(lockedTarget.percTarget.getAgentNumber(), xPos, yPos, direction,
                            lockedTarget.percTarget.getXPos(), lockedTarget.percTarget.getYPos(), lockedTarget.percTarget.getDirection(),
                            tmpTWAt.relXPosTar, tmpTWAt.relYPosTar));
 
 //           while(epiPos.size() > 0) 
 //               ((AgTarPos)epiPos.removeFirst()).printPos();
            // Here the update procedure must take place and the previous episode list be purged.
 //           epiPos.clear();
            episodeStatus=0;
            updateQStatus=2;
        }
     }
     else {
         System.out.println("Notification of Interception, but no target found!!! \n Exiting...");
         System.exit(1);
     }    
}

/************************************************  updateQ *****************************************/
//

public void updateQ() {
    
    switch(updateQStatus) {
                case 0 : 
                    // System.out.println("Nu in updateQ met updateQStatus=0");
                    break;
                
                case 1 : {
                    // System.out.println("Nu in updateQ met updateQStatus=1");

                    if(epiPos.size() >= 2){
                        AgTarPos posTmin1=((AgTarPos)epiPos.get(epiPos.size()-2));
                        AgTarPos posT=((AgTarPos)epiPos.get(epiPos.size()-1));
                        if(posTmin1 != null){                                
                                rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar]+=
                                    -1 + rlGrid[posT.dirAg][posT.relXPosTar][posT.relYPosTar][posT.dirTar] -
                                    rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar];
                                    
                        }
                        else {
                            System.out.println("Nu in updateQ with status=2, maar posTmin1=null; exiting");
                            System.exit(1);
                        }

                    }
                    updateQStatus=0;
                };
                break;
                
                case 2 :{

                    if(epiPos.size() >= 2){
                        AgTarPos posTmin1=((AgTarPos)epiPos.get(epiPos.size()-2));
                        AgTarPos posT=((AgTarPos)epiPos.get(epiPos.size()-1));
                        if(posTmin1 != null){
                                rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar]=0;
                                rlGrid[posT.dirAg][posT.relXPosTar][posT.relYPosTar][posT.dirTar]=0;
                        }
                        else {
                            System.exit(1);
                        }
                    }

                    while(epiPos.size() > 0) 
                      // Here the update procedure must take place and the previous episode list be purged.
                    epiPos.clear();
                    updateQStatus=0;
                };
                    break;
    
                case 3 : {
                    if(epiPos.size() > 1){
                        AgTarPos posTmin1=((AgTarPos)epiPos.get(epiPos.size()-2));
                        AgTarPos posT=((AgTarPos)epiPos.get(epiPos.size()-1));
                        if(posTmin1 != null){
                          
                                rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar]+=
                                    -1 + rlGrid[posT.dirAg][posT.relXPosTar][posT.relYPosTar][posT.dirTar] -
                                    rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar];
                                    
                                rlGrid[posT.dirAg][posT.relXPosTar][posT.relYPosTar][posT.dirTar] -= 1;    

                              //  System.out.println("Q_t-1 na update: " +  rlGrid[posTmin1.dirAg][posTmin1.relXPosTar][posTmin1.relYPosTar][posTmin1.dirTar]);   
                                    
                                while(epiPos.size() > 0) 
                                    epiPos.removeFirst();
                        }
                        else {
                            System.out.println("Nu in updateQ with status=3, maar posTmin1=null; exiting");
                            System.exit(1);
                        }

                    }
                    else if (epiPos.size()==1){  // De Target is maar n keer gezien en meteen verdwenen
                        AgTarPos posT=(AgTarPos)epiPos.getLast();
                        rlGrid[posT.dirAg][posT.relXPosTar][posT.relYPosTar][posT.dirTar] -= 1;
                        while(epiPos.size() > 0) 
                            epiPos.removeFirst();
                        
                    }

            };
            break;
    };
            
    updateQStatus=0;
}

/************************************************  ReportIfEaten ***********************************/
// Method to report if the bug has eaten.

public boolean reportIfEaten() {
    // if ( haveEaten )
       // System.out.println("Bug " + agentNumber + " has found food at " + xPos + ", " + yPos);
        return haveEaten;
    }
	//{{DECLARE_CONTROLS
	//}}


  public Object setBugColor(byte c) { 
    bugColor = c; 
    return this; 
  }

  public Object drawSelfOn (Raster r) {
    r.drawPointX$Y$Color (xPos, yPos, bugColor);
    return this;
  }

}